-- SRB2Kart Lua Script
-- Last frame variable tracking.

-- LastFrame_Define (def, key...)
-- Register the definition under keys in order. The registered defintion can
-- then be accessed via LastFrames. So with the following keys:
--   "type", "state"
-- ...you would access via:
--   LastFrames.type.state
-- Please use the SRB2 typedefs as userdata definition key.

-- LastFrame_Spawn (dest, regdef)
-- Call this function on initialization of a table or userdata to copy a
-- registered definition to it.

-- LastFrame_Copy (dest, regdef)
-- Call this function on script load to copy a registered definition to an
-- already existing table or userdata.

-- LastFrame (dest)
-- Perform last frame variable copying on a table or userdata. Call this
-- function after having copied a registered definition. Call this function
-- consistently, once you don't need to access the last frame variables
-- anymore.

	if LastFrame == nil then
--

local TABLE = "table"

local
function NilTable (u, k)
	if u[k] == nil
	then
		u[k] = {}
	end
	return u[k]
end

local
function Iterm (t, kfn, vfn, a)
	for k,v in pairs (t)
	do
		if type (v) == TABLE
		then
			Iterm(v, kfn, vfn, { kfn(k, a) })
		else
			vfn(k, v, a)
		end
	end
end

-- spaghetti

-- { m }

local
function Ikdef (k, a)
	return NilTable(a[1], k)
end

local
function Ivdef (k, v, a)
	a[1][v] = true
end

-- { u, d }

local
function Ikcpy (k, a)
	return a[1][k], NilTable(a[2], k)
end

local
function Ikdes (k, a)
	return a[1][k], a[2][k]
end

local
function Ivset (k, v, a)
	a[2][k] = a[1][k]
end

rawset (_G, "LastFrames", {})

rawset (_G, "LastFrame_Define",
function (t, ...)
	local m = LastFrames
	local a = { ... }
	for _,v in pairs (a)
	do
		m = NilTable($, v)
	end
	Iterm(t, Ikdef, Ivdef, { m })
end)

rawset (_G, "LastFrame_Spawn",
function (u, s)
	if u._lastframe == nil
	then
		LastFrame_Copy(u, s)
	end
end)

rawset (_G, "LastFrame_Copy",
function (u, s)
	u._lastframe = s
	if u._lastframe_c == nil
	then
		u._lastframe_c = 1
	else
		u._lastframe_c = $+1
	end
	if u._lastframe_n == nil
	then
		u._lastframe_n = 0
	end
	Iterm(u._lastframe, Ikcpy, Ivset, { u, NilTable(u,"lastframe") })
end)

rawset (_G, "LastFrame",
function (u)
	u._lastframe_n = $+1
	if (u._lastframe_n == u._lastframe_c)
	then
		Iterm(u._lastframe, Ikdes, Ivset, { u, u.lastframe })
		u._lastframe_n = 0
	end
end)

	end
