-- Depends: Queue, LastFrame

	if Rotate == nil then
--

rawset (_G, "Rotate",
{
	COUNT_RACES_FINISH = 0,
	COUNT_RACES_EXIT   = 1,
})

local TICMINUTE  = 60*TICRATE

local inter      = false
local intertic   = 0
local mapload    = false

local
function Plural (s, n)
	if (n > 1)
	then
		return s .. "s"
	else
		return s
	end
end

local
function Rotatetime (minutes, races)
	if (minutes < 1 and races < 1)
	then
		return nil
	end

	local s

	if (minutes > 0)
	then
		s = "after ".. minutes .. Plural(" minute", minutes)
	end

	if (races > 0)
	then
		local racecount

		if (Rotate.cv_rotate_count_races.value == Rotate.COUNT_RACES_FINISH)
		then
			racecount = " full race"
		else
			racecount = " map exit"
		end

		if (minutes > 0)
		then
			s = $.." and a minimum of "
		else
			s = "after "
		end

		s = $.. races .. Plural(racecount, races)
	end

	return s .. ", upon tally"
end

local
function Rotation ()
	return (
	Rotate.cv_rotate.value and (
	Rotate.cv_rotate_minutes.value > 0 or
	Rotate.cv_rotate_races.value > 0
	))
end

local
function InitRotate (p)
	p.Rotate =
	{
		mapsplayed   = 0,
		playtime     = 0,
		exceptions   = 0,
		exempt       = false,
	}
end

local
function NilRotate (p)
	local rot = p.Rotate
	if rot == nil
	then
		InitRotate(p)
	else
		return rot
	end
end

local
function ResetRotate (p)
	local rot = p.Rotate
	if rot ~= nil
	then
		rot.mapsplayed = 0
		rot.playtime   = 0
		rot.exceptions = 0
	end
end

local
function Rotatersiter (addmaps)
	local longest
	local longestrot

	local waiting = false

	local rot

	for p in players.iterate
	do
		if (p.spectatewait)
		then
			waiting = true
			break
		end
	end

	for p in players.iterate
	do
		rot = p.Rotate
		if rot ~= nil and -- very edge case of player joining right now (?)
		(not p.spectator and not p.lastframe.spectator) -- anti cheese
		then
			if (addmaps)
			then
				rot.mapsplayed = $+1
			end

			if (waiting and not rot.exempt and ( longest == nil or (
				rot.mapsplayed >= longestrot.mapsplayed and
				rot.playtime   >= longestrot.playtime) ))
			then
				longest    = p
				longestrot = rot -- could be wrong??
			end
		end
	end

	return longest, longestrot
end

local
function Nextmapchanges ()
	local when

	if (Rotate.cv_rotate.value)
	then
		when = Rotatetime(
		Rotate.cv_rotate_minutes.value,
		Rotate.cv_rotate_races.value)
	end

	if when == nil
	then
		print ("Racers will not be rotated out.")
	else
		local place
		place = Rotate.cv_rotate_max_place.value
		if (place > 1)
		then
			when = $.. ", except if they place ".. ( place - 1 ) .." or higher"
		end
		print ("Racers will be rotated out ".. when .. ".")
	end
end

local cvars =
{
	rotate               = { "Yes", CV_YesNo },
	rotate_minutes       = { "10",  CV_Unsigned },
	rotate_races         = {  "4",  CV_Unsigned },
	rotate_count_races   = { "Finish",
	{
		Finish = Rotate.COUNT_RACES_FINISH,
		Exit   = Rotate.COUNT_RACES_EXIT,
	}},
	rotate_max_place     = {  "1",
	{
		MIN    =  1,
		MAX    = 16,
	}},
	rotate_max_exception = {  "1", CV_Unsigned },
}

for k,v in pairs (cvars)
do
	Rotate["cv_" .. k] = CV_RegisterVar(
	{
		k, v[1],
		CV_SAVE|CV_NETVAR|CV_CALL|CV_NOINIT, v[2],
		Nextmapchanges
	})
end

cvars = nil

COM_AddCommand("rotate_exempt", function (p, no)
	if no == nil
	then
		CONS_Printf(p,
		"rotate_exempt <playernum>: Toggle exemption from rotation.")
	else
		no = tonumber($)
		if (no < 0 or no > 15)
		then
			CONS_Printf(p,
			"Player number out of range.")
		else
			local t = players[no]
			if t == nil
			then
				CONS_Printf(p,
				"Player not in game.")
			else
				NilRotate(t)
				local how
				if (t.Rotate.exempt)
				then
					t.Rotate.exempt = false
					how = "no longer"
				else
					t.Rotate.exempt = true
					how = "now"
				end

				print (players[no].name .." is ".. how .." exempt from rotation.")
			end
		end
	end
end, 1)

COM_AddCommand("rotate_exemptions", function (p)
	local rot
	for p in players.iterate
	do
		rot = p.Rotate
		if rot ~= nil and (rot.exempt)
		then
			CONS_Printf(p, p.name)
		end
	end
end)

LastFrame_Define(
{
	"spectator",
	kartstuff =
	{
		k_position,
	},
}, "player_t")

addHook ("PlayerSpawn", function (p)
	if (p.jointime == 0)
	then
		LastFrame_Spawn(p, LastFrames.player_t)

		InitRotate(p)
	end
end)

for p in players.iterate
do
	LastFrame_Copy(p, LastFrames.player_t)
end

addHook ("MobjThinker", function (mo)
	local p = mo.player
	if p ~= nil -- some fuckers spawn MT_PLAYER
	then
		local rot = p.Rotate
		if rot ~= nil and (not p.spectator)
		then
			rot.playtime = $+1
		end

		LastFrame(p)
	end
end, MT_PLAYER)

addHook ("ThinkFrame", function ()
	-- Delayed here because hudlines is cleared on map load.
	if (mapload)
	then
		if (Rotation())
		then
			local p, rot = Rotatersiter(false)
			if p ~= nil
			then
				local when = Rotatetime(
				Rotate.cv_rotate_minutes.value - rot.playtime / TICMINUTE,
				Rotate.cv_rotate_races.value - rot.mapsplayed)
				if when == nil
				then
					when = "at the end of this race" -- I am to be lazy at times
				end
				if (
					Rotate.cv_rotate_max_place.value > 1 and
					rot.exceptions < Rotate.cv_rotate_max_exception.value
					)
				then
					when = $.. ", except if you place "..
					( Rotate.cv_rotate_max_place.value - 1 ) .." or higher"
				end
				CONS_Printf(p,
				"\4*" .. "You will be rotated out ".. when .. ".")
			end
		end
		mapload = false
	end
end)

addHook ("IntermissionThinker", function ()
	if (inter or not Rotation()) then return end
	intertic = $ + 1
	if (intertic < 2) then return end
	inter = true

	if (Rotate.cv_rotate_count_races.value == Rotate.COUNT_RACES_FINISH)
	then
		local finish = false
		for p in players.iterate
		do
			if (p.exiting)
			then
				finish = true
				break
			end
		end
		if (not finish)
		then
			return
		end
	end

	local p, rot = Rotatersiter(true)
	if p ~= nil
	then
		if (
			p.lastframe.kartstuff[k_position] <
			Rotate.cv_rotate_max_place.value and

			rot.exceptions < Rotate.cv_rotate_max_exception.value
			)
		then
			rot.exceptions = $+1
		elseif ((
			Rotate.cv_rotate_minutes.value < 1 or
			rot.playtime / TICMINUTE >= Rotate.cv_rotate_minutes.value
			) and
			(
			Rotate.cv_rotate_races.value < 1 or
			rot.mapsplayed >= Rotate.cv_rotate_races.value
			))
		then
			p.spectator = true
			print ("\4*" .. p.name .." has been rotated out")

			for p in players.iterate
			do
				ResetRotate(p)
			end
		end
	end
end)

addHook ("MapLoad", function ()
	inter    = false
	intertic = 0
	mapload  = true
end)

	end
